/* Arquivo              : Timer.c        */
/* Autores              : Lanza - Rainer */
/* Modificacoes         : Fernando Barros*/
/* Modificacoes para pmp em rede: Marcos Tadeu  */
/* Data Inicio          : 29-04-91       */
/* Ult. mod. 12-12-92                    */
/*           24-11-93 MT: Prim_Fila_Livres = 1, em vez de zero em iniciar */
/*                        Timer_Tick                                      */
/*           25-11-93 MT: Colocado #ifdef NO_TIMER para retorno imediato  */
/*                        nas funcoes do timer, para utilizacao durante a */
/*                        fase de debug.                                  */
/*                        Obter_Timer_Estourado retorna 0 se nao houver timer */
/*           06-12-93 MT: Vetor de interrupcao restaurado e novamente colocado*/
/*                        para Interruption_Timer a cada inicio e fim de  */
/*                        rotina do Timer.                                */

/**********************************************************/
/*                                                        */
/*        Definicao das funcoes de temporizacao           */
/*                                                        */
/**********************************************************/

/***  Versao do artigo ***/

#include <dos.h>

#include <conio.h> //cprintf

#include <cntglob.h>
#include <fncesc.h>


/* Constantes */

#ifdef MPMP
 #include <CntMPDL.h>  /* Num_Estacoes */

 #define Tam_Max_Buffer_Saida       (Num_Estacoes + 8) /* 7 (TR,TRQ,...) + Num_Estacoes*TI */
 #define Nr_Max_Descritor_Tempo     (Num_Estacoes + 8)

 extern char far *video;

 void  Tratar_Temp_Estourado (void);
#endif

#ifdef PMP_REDE
 #include <cntdl_tx.h> /* Definicao da constante Num_Receptores */

 #define Tam_Max_Buffer_Saida       (Num_Receptores + 2)
 #define Nr_Max_Descritor_Tempo     Tam_Max_Buffer_Saida*(Num_Receptores + 2)

 void  Tratar_Temp_Estourado (void);
#endif

#ifdef SIMUL_PMP
 #include <cntdl_tx.h> /* Definicao da constante Num_Receptores */

 #define Tam_Max_Buffer_Saida       18 /*(Num_Receptores + 2)*/
 #define Nr_Max_Descritor_Tempo     Tam_Max_Buffer_Saida*18 /*(Num_Receptores + 2)*/

 void  Tratar_Temp_Estourado_Tx (void);
 void  Tratar_Temp_Estourado_Rx (unsigned char);

#endif


#ifdef  SIMUL_ROT
 #include <cntmdl.h>

 #define Tam_Max_Buffer_Saida       4
 #define Nr_Max_Descritor_Tempo     Tam_Max_Buffer_Saida*(Num_Estacoes + 1)

 void  Tratar_Temp_Estourado (unsigned char);
#endif


#ifdef DIFUSAO_SIMPLES
 #define Tam_Max_Buffer_Saida       4
 #define Nr_Max_Descritor_Tempo     Tam_Max_Buffer_Saida

 void  Tratar_Temp_Estourado (void);
#endif


/* Se nao for o simulador */
/* #define Nr_Max_Descritor_Tempo    Tam_Max_Buffer_Saida */

#define Nr_Max_Intervalos_Tempo    1000 /* 6144 */


/* Tipos */

typedef struct Tipo_Descritor_Tempo
  {
    unsigned int         Proximo;
    unsigned char        Sendo_Utilizado;
    unsigned char        Conexao;
    unsigned int         Posicao_Temporizacao;
  } Tipo_Descritor_Tempo;


typedef struct Tipo_Buffer_Saida
  {
    unsigned char       Buffer[Tam_Max_Buffer_Saida];
    unsigned int        In;
    unsigned int        Out;
  } Tipo_Buffer_Saida;


/* Variaveis */

unsigned int            Temporizacao [Nr_Max_Intervalos_Tempo];
unsigned int            Contador_Tempo;
Tipo_Descritor_Tempo    Descritor_Tempo [Nr_Max_Descritor_Tempo];
unsigned int            Prim_Fila_Livres;
unsigned int            Prim_Fila_Limbo [Nr_Max_Conexoes];
unsigned int            Timer_Estourado [Nr_Max_Conexoes];
Tipo_Buffer_Saida       Buffer_Saida [Nr_Max_Conexoes];
void interrupt          (*Interrupcao_Original)();


/* Cabecalho das funcoes */

void Inicializar_Timer_Tick (void);
void Finalizar_Timer_Tick (void);
void Criar_Timer (unsigned char,unsigned int,unsigned int *);
unsigned char Reiniciar_Timer (unsigned int,unsigned int);
unsigned int Finalizar_Timer (unsigned int *);
unsigned int Obter_Timer_Estourado (unsigned char);
void interrupt Interrupcao_Timer (void);


/****************************************************************************/

void Inicializar_Timer_Tick (void)
  {

   unsigned int i;

   Prim_Fila_Livres = 1; // MT: Causava ERRO: todos os codigos de tempo
                         //     sao iniciados com 0. Com isto, ao se obter
                         //     o timer estourado que usava o descritor 0
                         //     todos os Cod_Temp_... tinham o valor 0.

   Contador_Tempo = 0;

   for (i=0; i<Nr_Max_Intervalos_Tempo; i++)
     Temporizacao [i] = Nr_Max_Descritor_Tempo;

   for (i=0; i<Nr_Max_Conexoes; i++)
     {
      Timer_Estourado[i] = 0;
      Buffer_Saida[i].In = 0;
      Buffer_Saida[i].Out = 0;
      Prim_Fila_Limbo[i] = Nr_Max_Descritor_Tempo;
     } /* for */

   for (i=0; i<Nr_Max_Descritor_Tempo; i++)
     {
      Descritor_Tempo[i].Proximo = i+1;
      Descritor_Tempo[i].Sendo_Utilizado = 0;
      Descritor_Tempo[i].Posicao_Temporizacao = Nr_Max_Intervalos_Tempo;
     } /* for */

  disable();
  Interrupcao_Original = getvect(0x1C);
  enable();

  } /* Inicializar_Timer_Tick */


void Inicializar_Timer_Tick2 (void)
  {
  #ifndef NO_TIMER
  disable();
  setvect(0x1C,Interrupcao_Timer);
  enable();
  #endif
  }

/****************************************************************************/

void Finalizar_Timer_Tick (void)
  {

   disable();
   setvect(0x1C,Interrupcao_Original);
   enable();

  } /* Finalizar_Timer_Tick */


/****************************************************************************/

void Criar_Timer (unsigned char Nr_Conexao,unsigned int Tempo,unsigned int *Cod_Temp)
  {
    unsigned int        Descritor_Temporario ;
    unsigned int        Temporizacao_Temporaria;
    unsigned int        Posicao_Temporaria;

    #ifdef NO_TIMER
    return (0);
    #endif

    disable ();
//  setvect(0x1C,Interrupcao_Original);
//  enable();

    if (Nr_Conexao < Nr_Max_Conexoes)
      {
      if (*Cod_Temp)
        {
        if (!Reiniciar_Timer(*Cod_Temp,Tempo))
          {
          *Cod_Temp=0;
          Criar_Timer (Nr_Conexao,Tempo,Cod_Temp);
          }
        }
      else
        {
        Descritor_Temporario  = Prim_Fila_Livres;
        Prim_Fila_Livres = Descritor_Tempo [Prim_Fila_Livres].Proximo;

        Posicao_Temporaria = (Contador_Tempo + Tempo) % Nr_Max_Intervalos_Tempo;
        Descritor_Tempo [Descritor_Temporario].Proximo = Nr_Max_Descritor_Tempo;
        Descritor_Tempo [Descritor_Temporario].Sendo_Utilizado = 1;
        Descritor_Tempo [Descritor_Temporario].Conexao = Nr_Conexao;
        Descritor_Tempo [Descritor_Temporario].Posicao_Temporizacao =
          Posicao_Temporaria;

        Temporizacao_Temporaria = Temporizacao [Posicao_Temporaria];
        if (Temporizacao_Temporaria == Nr_Max_Descritor_Tempo)
          Temporizacao [Posicao_Temporaria] = Descritor_Temporario;
        else
          {
          while (Descritor_Tempo [Temporizacao_Temporaria].Proximo != Nr_Max_Descritor_Tempo)
            Temporizacao_Temporaria = Descritor_Tempo [Temporizacao_Temporaria].Proximo;
          Descritor_Tempo [Temporizacao_Temporaria].Proximo = Descritor_Temporario ;
          }

//        disable();
//        setvect(0x1C,Interrupcao_Timer);
        *Cod_Temp=Descritor_Temporario;
//        enable ();
//        return (Descritor_Temporario);
        }
      }
    else
//      {
//        disable();
//        setvect(0x1C,Interrupcao_Timer);
        *Cod_Temp=0;
//        enable ();
//        return (Nr_Max_Descritor_Tempo);
//      }
  enable();
  } /* Criar_Timer */


/****************************************************************************/

unsigned int Finalizar_Timer (unsigned int *Add_Nr_Temporizador)
  {
  unsigned int         Descritor_Atual;
  unsigned int Nr_Temporizador,sss;

  #ifdef NO_TIMER
  return (0);
  #endif

  disable ();
//    setvect(0x1C,Interrupcao_Original);
//    enable();

  Nr_Temporizador=*Add_Nr_Temporizador;
  *Add_Nr_Temporizador=0;
  if ((Nr_Temporizador < Nr_Max_Descritor_Tempo) &&
      (Descritor_Tempo [Nr_Temporizador].Sendo_Utilizado))
    {
    if (Temporizacao [Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao] ==
        Nr_Temporizador)
      Temporizacao [Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao] =
        Descritor_Tempo [Nr_Temporizador].Proximo;
    else
      {
      Descritor_Atual =
          Temporizacao [Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao];
      while (Descritor_Tempo [Descritor_Atual].Proximo != Nr_Temporizador)
        Descritor_Atual = Descritor_Tempo [Descritor_Atual].Proximo;
      Descritor_Tempo [Descritor_Atual].Proximo =
          Descritor_Tempo [Nr_Temporizador].Proximo;
      }
    Descritor_Tempo [Nr_Temporizador].Proximo = Prim_Fila_Livres;
    Descritor_Tempo [Nr_Temporizador].Sendo_Utilizado = 0;
    Prim_Fila_Livres = Nr_Temporizador;
//        disable();
//        setvect(0x1C,Interrupcao_Timer);
    enable ();
    return (Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao - Contador_Tempo);
    }
  else
    {
//        disable();
//        setvect(0x1C,Interrupcao_Timer);

    if (Buffer_Saida[0].Out!=Buffer_Saida[0].In)
      {
      sss=Buffer_Saida[0].Out;
      while ((sss!=Buffer_Saida[0].In) && (Buffer_Saida[0].Buffer[sss]!=Nr_Temporizador))
        sss=(sss+1) % Tam_Max_Buffer_Saida;
      if (Buffer_Saida[0].Buffer[sss]==Nr_Temporizador)
        Buffer_Saida[0].Buffer[sss]=0;
//        cprintf("-Fin-");
      }

    enable ();
    return (0);
    }
  } /* Finalizar_Timer */


/****************************************************************************/

unsigned char Reiniciar_Timer (unsigned int Nr_Temporizador,unsigned int Tempo)
  {
  unsigned int         Temporizacao_Temporaria;
  unsigned int         Posicao_Temporaria;
  unsigned int         Descritor_Atual,sss;

  #ifdef NO_TIMER
  return (0);
  #endif

  disable ();
//    setvect(0x1C,Interrupcao_Original);
//    enable();
  if ((Nr_Temporizador < Nr_Max_Descritor_Tempo) &&
      (Descritor_Tempo [Nr_Temporizador].Sendo_Utilizado))
    {
    if (Temporizacao [Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao] ==
        Nr_Temporizador)
      Temporizacao [Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao] =
        Descritor_Tempo [Nr_Temporizador].Proximo;
    else
      {
      Descritor_Atual=
        Temporizacao [Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao];
      while (Descritor_Tempo [Descritor_Atual].Proximo != Nr_Temporizador)
        Descritor_Atual = Descritor_Tempo [Descritor_Atual].Proximo;
      Descritor_Tempo [Descritor_Atual].Proximo =
                    Descritor_Tempo [Nr_Temporizador].Proximo;
      }
    Descritor_Tempo [Nr_Temporizador].Proximo = Nr_Max_Descritor_Tempo;

    Posicao_Temporaria = (Contador_Tempo + Tempo) % Nr_Max_Intervalos_Tempo;
    Descritor_Tempo [Nr_Temporizador].Posicao_Temporizacao =
      Posicao_Temporaria;

    Temporizacao_Temporaria = Temporizacao [Posicao_Temporaria];
    if (Temporizacao_Temporaria == Nr_Max_Descritor_Tempo)
      Temporizacao [Posicao_Temporaria] = Nr_Temporizador;
    else
      {
      while (Descritor_Tempo [Temporizacao_Temporaria].Proximo != Nr_Max_Descritor_Tempo)
        {
        Temporizacao_Temporaria = Descritor_Tempo [Temporizacao_Temporaria].Proximo;
        }
      Descritor_Tempo [Temporizacao_Temporaria].Proximo = Nr_Temporizador;
      Descritor_Tempo [Nr_Temporizador].Proximo = Nr_Max_Descritor_Tempo;
      }

//    disable();
//    setvect(0x1C,Interrupcao_Timer);
    enable ();
    return (1);
    }
  else
    {
//    disable();
//    setvect(0x1C,Interrupcao_Timer);

    if (Buffer_Saida[0].Out!=Buffer_Saida[0].In)
      {
      sss=Buffer_Saida[0].Out;
      while ((sss!=Buffer_Saida[0].In) && (Buffer_Saida[0].Buffer[sss]!=Nr_Temporizador))
        sss=(sss+1) % Tam_Max_Buffer_Saida;
      if (Buffer_Saida[0].Buffer[sss]==Nr_Temporizador)
        Buffer_Saida[0].Buffer[sss]=0;
//        cprintf("-Rein-");
      }

    enable ();
    return (0);
    }
  } /* Reiniciar_Timer */


/****************************************************************************/

unsigned int Obter_Timer_Estourado (unsigned char Nr_Conexao)
  {
  unsigned int               Resultado;
  unsigned int               Descritor_Lido;

  #ifdef NO_TIMER
  return (0);
  #endif

  disable();
//   setvect(0x1C,Interrupcao_Original);
//   enable();

  if (Buffer_Saida[Nr_Conexao].In == Buffer_Saida[Nr_Conexao].Out)
    {
//      disable();
//      setvect(0x1C,Interrupcao_Timer);
    enable();
    return (0); // (Nr_Max_Descritor_Tempo);
    } /* if */
  else
    {
    Resultado = Buffer_Saida[Nr_Conexao].Buffer[Buffer_Saida[Nr_Conexao].Out];
    Timer_Estourado[Nr_Conexao]--;
    Buffer_Saida[Nr_Conexao].Out = Buffer_Saida[Nr_Conexao].Out + 1;
    if (Buffer_Saida[Nr_Conexao].Out == Tam_Max_Buffer_Saida)
      Buffer_Saida[Nr_Conexao].Out = 0;
    Descritor_Lido = Prim_Fila_Limbo[Nr_Conexao];
    Prim_Fila_Limbo[Nr_Conexao] = Descritor_Tempo[Prim_Fila_Limbo[Nr_Conexao]].Proximo;
    Descritor_Tempo[Descritor_Lido].Proximo = Prim_Fila_Livres;
    Prim_Fila_Livres = Descritor_Lido;
//    disable();
//    setvect(0x1C,Interrupcao_Timer);
    enable();
    return(Resultado);
    } /* else */

  } /* Obter_Timer_Estourado */


/****************************************************/

void interrupt Interrupcao_Timer (void)
  {

  unsigned int          Descritor_Estourado;
  unsigned int          Descritor_Limbo_Atual;
  unsigned char         Conexao;

  Contador_Tempo = ++Contador_Tempo % Nr_Max_Intervalos_Tempo;
  while (Temporizacao [Contador_Tempo] != Nr_Max_Descritor_Tempo)
    {
    Descritor_Estourado = Temporizacao [Contador_Tempo];
    Temporizacao[Contador_Tempo] = Descritor_Tempo[Descritor_Estourado].Proximo;
    Descritor_Tempo [Descritor_Estourado].Sendo_Utilizado = 0;
    Conexao = Descritor_Tempo [Descritor_Estourado].Conexao;
    Buffer_Saida [Conexao].Buffer [Buffer_Saida [Conexao].In] = Descritor_Estourado;
    Buffer_Saida [Conexao].In = (Buffer_Saida [Conexao].In + 1) % Tam_Max_Buffer_Saida;
    Descritor_Limbo_Atual = Prim_Fila_Limbo [Conexao];
    if (Descritor_Limbo_Atual == Nr_Max_Descritor_Tempo)
      {
      Prim_Fila_Limbo [Conexao] = Descritor_Estourado;
      Descritor_Tempo [Descritor_Estourado].Proximo = Nr_Max_Descritor_Tempo;
      } /* if */
    else
      {
      while (Descritor_Tempo[Descritor_Limbo_Atual].Proximo != Nr_Max_Descritor_Tempo)
        {
        Descritor_Limbo_Atual = Descritor_Tempo [Descritor_Limbo_Atual].Proximo;
        } /* while */
      Descritor_Tempo [Descritor_Limbo_Atual].Proximo = Descritor_Estourado;
      Descritor_Tempo [Descritor_Estourado].Proximo = Nr_Max_Descritor_Tempo;
      } /* else */
    Timer_Estourado [Conexao]++;

/*************************************************************************


   se for para o RM-OSI



    if (Conexao < Nr_Max_Conexoes_DLPP)
#ifdef T_ENLACE
      Enviar_Rotina_Fila_Gerente_RMOSI (Executar_DL_Temporizador,Conexao)
#endif
;
    else if (Conexao < Nr_Max_Conexoes_DLPP + Nr_Max_Conexoes_Transporte)
#ifdef T_TRANSPORTE
      Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Timer_Transporte,Conexao - Nr_Max_Conexoes_DLPP)
#endif
;
    else
#ifdef T_SESSAO
      Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Timer_Sessao,Conexao - Nr_Max_Conexoes_DLPP - Nr_Max_Conexoes_Transporte)
#endif
;


*************************************************************************/

   /********* ponto a multi-ponto na rede ****/

#ifdef PMP_REDE

    Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado,0);

#endif



    /*********    Se for o simulador   *********/
#ifdef SIMUL_PMP

    if (Conexao == 0)
      Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado_Tx,0);
    else
      Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado_Rx,Conexao);
#endif

#ifdef SIMUL_ROT

    Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado,Conexao);

#endif

#ifdef DIFUSAO_SIMPLES

    Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado,0);

#endif

    /************  Multiponto a Multiponto *************/
#ifdef MPMP

 Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado,0);

#endif

       /*********   Se nao for o simulador   **************************/
       /*                                                             */
       /* Enviar_Rotina_Fila_Gerente_RMOSI (Tratar_Temp_Estourado,0); */
       /*                                                             */
       /***************************************************************/

      } /* while */
  } /* Interrupcao_Timer */
